"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const ConfDAL_1 = require("./fileDALs/ConfDAL");
const StatDAL_1 = require("./fileDALs/StatDAL");
const ConfDTO_1 = require("../dto/ConfDTO");
const TransactionDTO_1 = require("../dto/TransactionDTO");
const DBBlock_1 = require("../db/DBBlock");
const MerkleDTO_1 = require("../dto/MerkleDTO");
const constants_1 = require("../common-libs/constants");
const PowDAL_1 = require("./fileDALs/PowDAL");
const fs = require('fs');
const path = require('path');
const readline = require('readline');
const _ = require('underscore');
const indexer = require('../indexer').Indexer;
const logger = require('../logger').NewLogger('filedal');
const constants = require('../constants');
class FileDAL {
    constructor(params) {
        this.setRevocating = (existing, revocation_sig) => {
            existing.revocation_sig = revocation_sig;
            existing.revoked = false;
            return this.idtyDAL.saveIdentity(existing);
        };
        this.rootPath = params.home;
        this.myFS = params.fs;
        this.sqliteDriver = params.dbf();
        this.wotb = params.wotb;
        this.profile = 'DAL';
        // DALs
        this.powDAL = new PowDAL_1.PowDAL(this.rootPath, this.myFS);
        this.confDAL = new ConfDAL_1.ConfDAL(this.rootPath, this.myFS);
        this.metaDAL = new (require('./sqliteDAL/MetaDAL').MetaDAL)(this.sqliteDriver);
        this.peerDAL = new (require('./sqliteDAL/PeerDAL').PeerDAL)(this.sqliteDriver);
        this.blockDAL = new (require('./sqliteDAL/BlockDAL').BlockDAL)(this.sqliteDriver);
        this.txsDAL = new (require('./sqliteDAL/TxsDAL').TxsDAL)(this.sqliteDriver);
        this.statDAL = new StatDAL_1.StatDAL(this.rootPath, this.myFS);
        this.idtyDAL = new (require('./sqliteDAL/IdentityDAL').IdentityDAL)(this.sqliteDriver);
        this.certDAL = new (require('./sqliteDAL/CertDAL').CertDAL)(this.sqliteDriver);
        this.msDAL = new (require('./sqliteDAL/MembershipDAL').MembershipDAL)(this.sqliteDriver);
        this.walletDAL = new (require('./sqliteDAL/WalletDAL').WalletDAL)(this.sqliteDriver);
        this.bindexDAL = new (require('./sqliteDAL/index/BIndexDAL').BIndexDAL)(this.sqliteDriver);
        this.mindexDAL = new (require('./sqliteDAL/index/MIndexDAL').MIndexDAL)(this.sqliteDriver);
        this.iindexDAL = new (require('./sqliteDAL/index/IIndexDAL').IIndexDAL)(this.sqliteDriver);
        this.sindexDAL = new (require('./sqliteDAL/index/SIndexDAL').SIndexDAL)(this.sqliteDriver);
        this.cindexDAL = new (require('./sqliteDAL/index/CIndexDAL').CIndexDAL)(this.sqliteDriver);
        this.newDals = {
            'powDAL': this.powDAL,
            'metaDAL': this.metaDAL,
            'blockDAL': this.blockDAL,
            'certDAL': this.certDAL,
            'msDAL': this.msDAL,
            'idtyDAL': this.idtyDAL,
            'txsDAL': this.txsDAL,
            'peerDAL': this.peerDAL,
            'confDAL': this.confDAL,
            'statDAL': this.statDAL,
            'walletDAL': this.walletDAL,
            'bindexDAL': this.bindexDAL,
            'mindexDAL': this.mindexDAL,
            'iindexDAL': this.iindexDAL,
            'sindexDAL': this.sindexDAL,
            'cindexDAL': this.cindexDAL
        };
    }
    async init(conf) {
        const dalNames = _.keys(this.newDals);
        for (const dalName of dalNames) {
            const dal = this.newDals[dalName];
            await dal.init();
        }
        logger.debug("Upgrade database...");
        await this.metaDAL.upgradeDatabase(conf);
        // Update the maximum certifications count a member can issue into the C++ addon
        const currencyParams = await this.getParameters();
        if (currencyParams && currencyParams.sigStock !== undefined && currencyParams.sigStock !== null) {
            this.wotb.setMaxCert(currencyParams.sigStock);
        }
    }
    getDBVersion() {
        return this.metaDAL.getVersion();
    }
    writeFileOfBlock(block) {
        return this.blockDAL.saveBlock(block);
    }
    writeSideFileOfBlock(block) {
        return this.blockDAL.saveSideBlock(block);
    }
    listAllPeers() {
        return this.peerDAL.listAll();
    }
    async getPeer(pubkey) {
        try {
            return this.peerDAL.getPeer(pubkey);
        }
        catch (err) {
            throw Error('Unknown peer ' + pubkey);
        }
    }
    async getWS2Peers() {
        return this.peerDAL.getPeersWithEndpointsLike('WS2P');
    }
    async getBlock(number) {
        const block = await this.blockDAL.getBlock(number);
        return block || null;
    }
    getAbsoluteBlockByNumberAndHash(number, hash) {
        return this.blockDAL.getAbsoluteBlock(number, hash);
    }
    getAbsoluteBlockByBlockstamp(blockstamp) {
        if (!blockstamp)
            throw "Blockstamp is required to find the block";
        const sp = blockstamp.split('-');
        const number = parseInt(sp[0]);
        const hash = sp[1];
        return this.getAbsoluteBlockByNumberAndHash(number, hash);
    }
    getBlockByBlockstampOrNull(blockstamp) {
        if (!blockstamp)
            throw "Blockstamp is required to find the block";
        const sp = blockstamp.split('-');
        const number = parseInt(sp[0]);
        const hash = sp[1];
        return this.getBlockByNumberAndHashOrNull(number, hash);
    }
    getBlockByBlockstamp(blockstamp) {
        if (!blockstamp)
            throw "Blockstamp is required to find the block";
        const sp = blockstamp.split('-');
        const number = parseInt(sp[0]);
        const hash = sp[1];
        return this.getBlockByNumberAndHash(number, hash);
    }
    async getBlockByNumberAndHash(number, hash) {
        try {
            const block = await this.getBlock(number);
            if (!block || block.hash != hash)
                throw "Not found";
            else
                return block;
        }
        catch (err) {
            throw 'Block ' + [number, hash].join('-') + ' not found';
        }
    }
    async getBlockByNumberAndHashOrNull(number, hash) {
        try {
            return await this.getBlockByNumberAndHash(number, hash);
        }
        catch (e) {
            return null;
        }
    }
    async existsNonChainableLink(from, vHEAD_1, sigStock) {
        // Cert period rule
        const medianTime = vHEAD_1 ? vHEAD_1.medianTime : 0;
        const linksFrom = await this.cindexDAL.reducablesFrom(from);
        const unchainables = _.filter(linksFrom, (link) => link.chainable_on > medianTime);
        if (unchainables.length > 0)
            return true;
        // Max stock rule
        let activeLinks = _.filter(linksFrom, (link) => !link.expired_on);
        return activeLinks.length >= sigStock;
    }
    async getCurrentBlockOrNull() {
        let current = null;
        try {
            current = await this.getBlockCurrent();
        }
        catch (e) {
            if (e != constants.ERROR.BLOCK.NO_CURRENT_BLOCK) {
                throw e;
            }
        }
        return current;
    }
    getPromoted(number) {
        return this.getBlock(number);
    }
    // Block
    lastUDBlock() {
        return this.blockDAL.lastBlockWithDividend();
    }
    getRootBlock() {
        return this.getBlock(0);
    }
    getPotentialRootBlocks() {
        return this.blockDAL.getPotentialRoots();
    }
    lastBlockOfIssuer(issuer) {
        return this.blockDAL.lastBlockOfIssuer(issuer);
    }
    getCountOfPoW(issuer) {
        return this.blockDAL.getCountOfBlocksIssuedBy(issuer);
    }
    getBlocksBetween(start, end) {
        return Promise.resolve(this.blockDAL.getBlocks(Math.max(0, start), end));
    }
    getForkBlocksFollowing(current) {
        return this.blockDAL.getNextForkBlocks(current.number, current.hash);
    }
    getPotentialForkBlocks(numberStart, medianTimeStart, maxNumber) {
        return this.blockDAL.getPotentialForkBlocks(numberStart, medianTimeStart, maxNumber);
    }
    async getBlockCurrent() {
        const current = await this.blockDAL.getCurrent();
        if (!current)
            throw 'No current block';
        return current;
    }
    getValidLinksTo(to) {
        return this.cindexDAL.getValidLinksTo(to);
    }
    getAvailableSourcesByPubkey(pubkey) {
        return this.sindexDAL.getAvailableForPubkey(pubkey);
    }
    async getIdentityByHashOrNull(hash) {
        const pending = await this.idtyDAL.getByHash(hash);
        if (!pending) {
            return this.iindexDAL.getFromHash(hash);
        }
        return pending;
    }
    getMembers() {
        return this.iindexDAL.getMembers();
    }
    // TODO: this should definitely be reduced by removing fillInMembershipsOfIdentity
    async getWritten(pubkey) {
        try {
            return await this.fillInMembershipsOfIdentity(this.iindexDAL.getFromPubkey(pubkey));
        }
        catch (err) {
            logger.error(err);
            return null;
        }
    }
    async getWrittenIdtyByPubkey(pubkey) {
        const idty = await this.iindexDAL.getFromPubkey(pubkey);
        if (!idty) {
            return null;
        }
        const membership = await this.mindexDAL.getReducedMS(pubkey);
        idty.revoked_on = membership.revoked_on;
        return idty;
    }
    async getWrittenIdtyByUID(uid) {
        const idty = await this.iindexDAL.getFromUID(uid);
        if (!idty) {
            return null;
        }
        const membership = await this.mindexDAL.getReducedMS(idty.pub);
        idty.revoked_on = membership.revoked_on;
        return idty;
    }
    async fillInMembershipsOfIdentity(queryPromise) {
        try {
            const idty = await Promise.resolve(queryPromise);
            if (idty) {
                const mss = await this.msDAL.getMembershipsOfIssuer(idty.pubkey);
                const mssFromMindex = await this.mindexDAL.reducable(idty.pubkey);
                idty.memberships = mss.concat(mssFromMindex.map((ms) => {
                    const sp = ms.created_on.split('-');
                    return {
                        blockstamp: ms.created_on,
                        membership: ms.leaving ? 'OUT' : 'IN',
                        number: sp[0],
                        fpr: sp[1],
                        written_number: parseInt(ms.written_on)
                    };
                }));
                return idty;
            }
        }
        catch (err) {
            logger.error(err);
        }
        return null;
    }
    async findPeersWhoseHashIsIn(hashes) {
        const peers = await this.peerDAL.listAll();
        return _.chain(peers).filter((p) => hashes.indexOf(p.hash) !== -1).value();
    }
    getTxByHash(hash) {
        return this.txsDAL.getTX(hash);
    }
    removeTxByHash(hash) {
        return this.txsDAL.removeTX(hash);
    }
    getTransactionsPending(versionMin = 0) {
        return this.txsDAL.getAllPending(versionMin);
    }
    async getNonWritten(pubkey) {
        const pending = await this.idtyDAL.getPendingIdentities();
        return _.chain(pending).where({ pubkey: pubkey }).value();
    }
    async getRevocatingMembers() {
        const revoking = await this.idtyDAL.getToRevoke();
        const toRevoke = [];
        for (const pending of revoking) {
            const idty = await this.getWrittenIdtyByPubkey(pending.pubkey);
            if (!idty.revoked_on) {
                toRevoke.push(pending);
            }
        }
        return toRevoke;
    }
    getToBeKickedPubkeys() {
        return this.iindexDAL.getToBeKickedPubkeys();
    }
    getRevokedPubkeys() {
        return this.mindexDAL.getRevokedPubkeys();
    }
    async searchJustIdentities(search) {
        const pendings = await this.idtyDAL.searchThoseMatching(search);
        const writtens = await this.iindexDAL.searchThoseMatching(search);
        const nonPendings = _.filter(writtens, (w) => {
            return _.where(pendings, { pubkey: w.pub }).length == 0;
        });
        const found = pendings.concat(nonPendings.map((i) => {
            // Use the correct field
            i.pubkey = i.pub;
            return i;
        }));
        return await Promise.all(found.map(async (f) => {
            const ms = await this.mindexDAL.getReducedMS(f.pub);
            if (ms) {
                f.revoked_on = ms.revoked_on ? parseInt(ms.revoked_on) : null;
                f.revoked = !!f.revoked_on;
                f.revocation_sig = ms.revocation || null;
            }
            return f;
        }));
    }
    async certsToTarget(pub, hash) {
        const certs = await this.certDAL.getToTarget(hash);
        const links = await this.cindexDAL.getValidLinksTo(pub);
        let matching = certs;
        await Promise.all(links.map(async (entry) => {
            entry.from = entry.issuer;
            const wbt = entry.written_on.split('-');
            const blockNumber = parseInt(entry.created_on); // created_on field of `c_index` does not have the full blockstamp
            const basedBlock = await this.getBlock(blockNumber);
            entry.block = blockNumber;
            entry.block_number = blockNumber;
            entry.block_hash = basedBlock ? basedBlock.hash : null;
            entry.linked = true;
            entry.written_block = parseInt(wbt[0]);
            entry.written_hash = wbt[1];
            matching.push(entry);
        }));
        matching = _.sortBy(matching, (c) => -c.block);
        matching.reverse();
        return matching;
    }
    async certsFrom(pubkey) {
        const certs = await this.certDAL.getFromPubkeyCerts(pubkey);
        const links = await this.cindexDAL.getValidLinksFrom(pubkey);
        let matching = certs;
        await Promise.all(links.map(async (entry) => {
            const idty = await this.getWrittenIdtyByPubkey(entry.receiver);
            entry.from = entry.issuer;
            entry.to = entry.receiver;
            const cbt = entry.created_on.split('-');
            const wbt = entry.written_on.split('-');
            entry.block = parseInt(cbt[0]);
            entry.block_number = parseInt(cbt[0]);
            entry.block_hash = cbt[1];
            entry.target = idty.hash;
            entry.linked = true;
            entry.written_block = parseInt(wbt[0]);
            entry.written_hash = wbt[1];
            matching.push(entry);
        }));
        matching = _.sortBy(matching, (c) => -c.block);
        matching.reverse();
        return matching;
    }
    async isSentry(pubkey, conf) {
        const current = await this.getCurrentBlockOrNull();
        if (current) {
            const dSen = Math.ceil(Math.pow(current.membersCount, 1 / conf.stepMax));
            const linksFrom = await this.cindexDAL.getValidLinksFrom(pubkey);
            const linksTo = await this.cindexDAL.getValidLinksTo(pubkey);
            return linksFrom.length >= dSen && linksTo.length >= dSen;
        }
        return false;
    }
    async certsFindNew() {
        const certs = await this.certDAL.getNotLinked();
        return _.chain(certs).where({ linked: false }).sortBy((c) => -c.block).value();
    }
    async certsNotLinkedToTarget(hash) {
        const certs = await this.certDAL.getNotLinkedToTarget(hash);
        return _.chain(certs).sortBy((c) => -c.block).value();
    }
    async getMostRecentMembershipNumberForIssuer(issuer) {
        const mss = await this.msDAL.getMembershipsOfIssuer(issuer);
        const reduced = await this.mindexDAL.getReducedMS(issuer);
        let max = reduced ? parseInt(reduced.created_on) : -1;
        for (const ms of mss) {
            max = Math.max(ms.number, max);
        }
        return max;
    }
    async lastJoinOfIdentity(target) {
        let pending = await this.msDAL.getPendingINOfTarget(target);
        return _(pending).sortBy((ms) => -ms.number)[0];
    }
    async findNewcomers(blockMedianTime = 0) {
        const pending = await this.msDAL.getPendingIN();
        const mss = await Promise.all(pending.map(async (p) => {
            const reduced = await this.mindexDAL.getReducedMS(p.issuer);
            if (!reduced || !reduced.chainable_on || blockMedianTime >= reduced.chainable_on || blockMedianTime < constants.TIME_TO_TURN_ON_BRG_107) {
                return p;
            }
            return null;
        }));
        return _.chain(mss)
            .filter((ms) => ms)
            .sortBy((ms) => -ms.sigDate)
            .value();
    }
    async findLeavers() {
        const mss = await this.msDAL.getPendingOUT();
        return _.chain(mss).sortBy((ms) => -ms.sigDate).value();
    }
    existsNonReplayableLink(from, to) {
        return this.cindexDAL.existsNonReplayableLink(from, to);
    }
    getSource(identifier, pos) {
        return this.sindexDAL.getSource(identifier, pos);
    }
    async isMember(pubkey) {
        try {
            const idty = await this.iindexDAL.getFromPubkey(pubkey);
            if (idty === null) {
                return false;
            }
            return true;
        }
        catch (err) {
            return false;
        }
    }
    async isMemberAndNonLeaver(pubkey) {
        try {
            const idty = await this.iindexDAL.getFromPubkey(pubkey);
            if (idty && idty.member) {
                return !(await this.isLeaving(pubkey));
            }
            return false;
        }
        catch (err) {
            return false;
        }
    }
    async isLeaving(pubkey) {
        const ms = await this.mindexDAL.getReducedMS(pubkey);
        return (ms && ms.leaving) || false;
    }
    async existsCert(cert) {
        const existing = await this.certDAL.existsGivenCert(cert);
        if (existing)
            return existing;
        const existsLink = await this.cindexDAL.existsNonReplayableLink(cert.from, cert.to);
        return !!existsLink;
    }
    deleteCert(cert) {
        return this.certDAL.deleteCert(cert);
    }
    deleteMS(ms) {
        return this.msDAL.deleteMS(ms);
    }
    async setRevoked(pubkey) {
        const idty = await this.getWrittenIdtyByPubkey(pubkey);
        idty.revoked = true;
        return await this.idtyDAL.saveIdentity(idty);
    }
    async getPeerOrNull(pubkey) {
        let peer = null;
        try {
            peer = await this.getPeer(pubkey);
        }
        catch (e) {
            if (e != constants.ERROR.BLOCK.NO_CURRENT_BLOCK) {
                throw e;
            }
        }
        return peer;
    }
    async removePeerByPubkey(pubkey) {
        return this.peerDAL.removePeerByPubkey(pubkey);
    }
    async findAllPeersNEWUPBut(pubkeys) {
        const peers = await this.listAllPeers();
        return peers.filter((peer) => pubkeys.indexOf(peer.pubkey) == -1
            && ['UP'].indexOf(peer.status) !== -1);
    }
    async listAllPeersWithStatusNewUP() {
        const peers = await this.peerDAL.listAll();
        return _.chain(peers)
            .filter((p) => ['UP']
            .indexOf(p.status) !== -1).value();
    }
    async listAllPeersWithStatusNewUPWithtout(pub) {
        const peers = await this.peerDAL.listAll();
        return _.chain(peers).filter((p) => p.status == 'UP').filter((p) => p.pubkey !== pub).value();
    }
    async findPeers(pubkey) {
        try {
            const peer = await this.getPeer(pubkey);
            return [peer];
        }
        catch (err) {
            return [];
        }
    }
    async getRandomlyUPsWithout(pubkeys) {
        const peers = await this.listAllPeersWithStatusNewUP();
        return peers.filter((peer) => pubkeys.indexOf(peer.pubkey) == -1);
    }
    async setPeerUP(pubkey) {
        try {
            const p = await this.getPeer(pubkey);
            p.status = 'UP';
            p.first_down = null;
            p.last_try = null;
            return this.peerDAL.savePeer(p);
        }
        catch (err) {
            return null;
        }
    }
    async setPeerDown(pubkey) {
        try {
            // We do not set mirror peers as down (ex. of mirror: 'M1_HnFcSms8jzwngtVomTTnzudZx7SHUQY8sVE1y8yBmULk')
            if (!pubkey.match(/_/)) {
                const p = await this.getPeer(pubkey);
                if (p) {
                    const now = (new Date()).getTime();
                    p.status = 'DOWN';
                    if (!p.first_down) {
                        p.first_down = now;
                    }
                    p.last_try = now;
                    await this.peerDAL.savePeer(p);
                }
            }
        }
        catch (err) {
            throw err;
        }
    }
    async saveBlock(block) {
        const dbb = DBBlock_1.DBBlock.fromBlockDTO(block);
        dbb.wrong = false;
        await Promise.all([
            this.saveBlockInFile(dbb),
            this.saveTxsInFiles(block.transactions, block.number, block.medianTime)
        ]);
    }
    async generateIndexes(block, conf, index, HEAD) {
        // We need to recompute the indexes for block#0
        if (!index || !HEAD || HEAD.number == 0) {
            index = indexer.localIndex(block, conf);
            HEAD = await indexer.completeGlobalScope(block, conf, index, this);
        }
        let mindex = indexer.mindex(index);
        let iindex = indexer.iindex(index);
        let sindex = indexer.sindex(index);
        let cindex = indexer.cindex(index);
        sindex = sindex.concat(await indexer.ruleIndexGenDividend(HEAD, this));
        sindex = sindex.concat(await indexer.ruleIndexGarbageSmallAccounts(HEAD, sindex, this));
        cindex = cindex.concat(await indexer.ruleIndexGenCertificationExpiry(HEAD, this));
        mindex = mindex.concat(await indexer.ruleIndexGenMembershipExpiry(HEAD, this));
        iindex = iindex.concat(await indexer.ruleIndexGenExclusionByMembership(HEAD, mindex, this));
        iindex = iindex.concat(await indexer.ruleIndexGenExclusionByCertificatons(HEAD, cindex, iindex, conf, this));
        mindex = mindex.concat(await indexer.ruleIndexGenImplicitRevocation(HEAD, this));
        await indexer.ruleIndexCorrectMembershipExpiryDate(HEAD, mindex, this);
        await indexer.ruleIndexCorrectCertificationExpiryDate(HEAD, cindex, this);
        return { HEAD, mindex, iindex, sindex, cindex };
    }
    async updateWotbLinks(cindex) {
        for (const entry of cindex) {
            const from = await this.getWrittenIdtyByPubkey(entry.issuer);
            const to = await this.getWrittenIdtyByPubkey(entry.receiver);
            if (entry.op == constants_1.CommonConstants.IDX_CREATE) {
                this.wotb.addLink(from.wotb_id, to.wotb_id);
            }
            else {
                // Update = removal
                this.wotb.removeLink(from.wotb_id, to.wotb_id);
            }
        }
    }
    async trimIndexes(maxNumber) {
        await this.bindexDAL.trimBlocks(maxNumber);
        await this.iindexDAL.trimRecords(maxNumber);
        await this.mindexDAL.trimRecords(maxNumber);
        await this.cindexDAL.trimExpiredCerts(maxNumber);
        await this.sindexDAL.trimConsumedSource(maxNumber);
        return true;
    }
    async trimSandboxes(block) {
        await this.certDAL.trimExpiredCerts(block.medianTime);
        await this.msDAL.trimExpiredMemberships(block.medianTime);
        await this.idtyDAL.trimExpiredIdentities(block.medianTime);
        await this.txsDAL.trimExpiredNonWrittenTxs(block.medianTime - constants_1.CommonConstants.TX_WINDOW);
        return true;
    }
    savePendingMembership(ms) {
        return this.msDAL.savePendingMembership(ms);
    }
    async saveBlockInFile(block) {
        await this.writeFileOfBlock(block);
    }
    saveSideBlockInFile(block) {
        return this.writeSideFileOfBlock(block);
    }
    async saveTxsInFiles(txs, block_number, medianTime) {
        return Promise.all(txs.map(async (tx) => {
            const sp = tx.blockstamp.split('-');
            tx.blockstampTime = (await this.getBlockByNumberAndHash(parseInt(sp[0]), sp[1])).medianTime;
            const txEntity = TransactionDTO_1.TransactionDTO.fromJSONObject(tx);
            txEntity.computeAllHashes();
            return this.txsDAL.addLinked(TransactionDTO_1.TransactionDTO.fromJSONObject(txEntity), block_number, medianTime);
        }));
    }
    async merkleForPeers() {
        let peers = await this.listAllPeersWithStatusNewUP();
        const leaves = peers.map((peer) => peer.hash);
        const merkle = new MerkleDTO_1.MerkleDTO();
        merkle.initialize(leaves);
        return merkle;
    }
    removeAllSourcesOfBlock(blockstamp) {
        return this.sindexDAL.removeBlock(blockstamp);
    }
    updateTransactions(txs) {
        return this.txsDAL.insertBatchOfTxs(txs);
    }
    savePendingIdentity(idty) {
        return this.idtyDAL.saveIdentity(idty);
    }
    revokeIdentity(pubkey) {
        return this.idtyDAL.revokeIdentity(pubkey);
    }
    async removeUnWrittenWithPubkey(pubkey) {
        return await this.idtyDAL.removeUnWrittenWithPubkey(pubkey);
    }
    async removeUnWrittenWithUID(pubkey) {
        return await this.idtyDAL.removeUnWrittenWithUID(pubkey);
    }
    registerNewCertification(cert) {
        return this.certDAL.saveNewCertification(cert);
    }
    saveTransaction(tx) {
        return this.txsDAL.addPending(tx);
    }
    async getTransactionsHistory(pubkey) {
        const history = {
            sent: [],
            received: [],
            sending: [],
            receiving: [],
            pending: []
        };
        const res = await Promise.all([
            this.txsDAL.getLinkedWithIssuer(pubkey),
            this.txsDAL.getLinkedWithRecipient(pubkey),
            this.txsDAL.getPendingWithIssuer(pubkey),
            this.txsDAL.getPendingWithRecipient(pubkey)
        ]);
        history.sent = res[0] || [];
        history.received = res[1] || [];
        history.sending = res[2] || [];
        history.pending = res[3] || [];
        return history;
    }
    async getUDHistory(pubkey) {
        const sources = await this.sindexDAL.getUDSources(pubkey);
        return {
            history: sources.map((src) => _.extend({
                block_number: src.pos,
                time: src.written_time
            }, src))
        };
    }
    savePeer(peer) {
        return this.peerDAL.savePeer(peer);
    }
    async getUniqueIssuersBetween(start, end) {
        const current = await this.blockDAL.getCurrent();
        const firstBlock = Math.max(0, start);
        const lastBlock = Math.max(0, Math.min(current.number, end));
        const blocks = await this.blockDAL.getBlocks(firstBlock, lastBlock);
        return _.chain(blocks).pluck('issuer').uniq().value();
    }
    /**
     * Gets a range of entries for the last `start`th to the last `end`th HEAD entry.
     * @param start The starting entry number (min. 1)
     * @param end The ending entry (max. BINDEX length)
     * @param property If provided, transforms the range of entries into an array of the asked property.
     */
    async range(start, end, property) {
        const range = await this.bindexDAL.range(start, end);
        if (property) {
            // Filter on a particular property
            return range.map((b) => b[property]);
        }
        else {
            return range;
        }
    }
    /**
     * Get the last `n`th entry from the BINDEX.
     * @param n The entry number (min. 1).
     */
    head(n) {
        return this.bindexDAL.head(n);
    }
    /***********************
     *    CONFIGURATION
     **********************/
    getParameters() {
        return this.confDAL.getParameters();
    }
    async loadConf(overrideConf, defaultConf = false) {
        let conf = ConfDTO_1.ConfDTO.complete(overrideConf || {});
        if (!defaultConf) {
            const savedConf = await this.confDAL.loadConf();
            const savedProxyConf = _(savedConf.proxyConf).extend({});
            conf = _(savedConf).extend(overrideConf || {});
            if (overrideConf.proxiesConf !== undefined) { }
            else {
                conf.proxyConf = _(savedProxyConf).extend({});
            }
        }
        if (this.loadConfHook) {
            await this.loadConfHook(conf);
        }
        return conf;
    }
    async saveConf(confToSave) {
        // Save the conf in file
        let theConf = confToSave;
        if (this.saveConfHook) {
            theConf = await this.saveConfHook(theConf);
        }
        return this.confDAL.saveConf(theConf);
    }
    /***********************
     *     WALLETS
     **********************/
    async getWallet(conditions) {
        let wallet = await this.walletDAL.getWallet(conditions);
        if (!wallet) {
            wallet = { conditions, balance: 0 };
        }
        return wallet;
    }
    saveWallet(wallet) {
        return this.walletDAL.saveWallet(wallet);
    }
    /***********************
     *     STATISTICS
     **********************/
    loadStats() {
        return this.statDAL.loadStats();
    }
    getStat(name) {
        return this.statDAL.getStat(name);
    }
    pushStats(stats) {
        return this.statDAL.pushStats(stats);
    }
    async cleanCaches() {
        await _.values(this.newDals).map((dal) => dal.cleanCache && dal.cleanCache());
    }
    async close() {
        await _.values(this.newDals).map((dal) => dal.cleanCache && dal.cleanCache());
        return this.sqliteDriver.closeConnection();
    }
    async resetPeers() {
        this.peerDAL.removeAll();
        return await this.close();
    }
    getLogContent(linesQuantity) {
        return new Promise((resolve, reject) => {
            try {
                let lines = [], i = 0;
                const logPath = path.join(this.rootPath, 'duniter.log');
                const readStream = fs.createReadStream(logPath);
                readStream.on('error', (err) => reject(err));
                const lineReader = readline.createInterface({
                    input: readStream
                });
                lineReader.on('line', (line) => {
                    line = "\n" + line;
                    lines.push(line);
                    i++;
                    if (i >= linesQuantity)
                        lines.shift();
                });
                lineReader.on('close', () => resolve(lines));
                lineReader.on('error', (err) => reject(err));
            }
            catch (e) {
                reject(e);
            }
        });
    }
}
exports.FileDAL = FileDAL;
//# sourceMappingURL=fileDAL.js.map