"use strict";
// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
Object.defineProperty(exports, "__esModule", { value: true });
const AbstractSQLite_1 = require("./AbstractSQLite");
const SandBox_1 = require("./SandBox");
const _ = require('underscore');
const moment = require('moment');
const constants = require('../../constants');
class DBTx {
    static fromTransactionDTO(tx) {
        const dbTx = new DBTx();
        dbTx.hash = tx.hash;
        dbTx.locktime = tx.locktime;
        dbTx.version = tx.version;
        dbTx.currency = tx.currency;
        dbTx.blockstamp = tx.blockstamp;
        dbTx.blockstampTime = tx.blockstampTime;
        dbTx.comment = tx.comment || "";
        dbTx.inputs = tx.inputs;
        dbTx.unlocks = tx.unlocks;
        dbTx.outputs = tx.outputs;
        dbTx.issuers = tx.issuers;
        dbTx.signatures = tx.signatures;
        dbTx.recipients = tx.outputsAsRecipients();
        dbTx.written = false;
        dbTx.removed = false;
        dbTx.output_base = tx.output_base;
        dbTx.output_amount = tx.output_amount;
        return dbTx;
    }
    static setRecipients(txs) {
        // Each transaction must have a good "recipients" field for future searchs
        txs.forEach((tx) => tx.recipients = DBTx.outputs2recipients(tx));
    }
    static outputs2recipients(tx) {
        return tx.outputs.map(function (out) {
            const recipent = out.match('SIG\\((.*)\\)');
            return (recipent && recipent[1]) || 'UNKNOWN';
        });
    }
}
exports.DBTx = DBTx;
class TxsDAL extends AbstractSQLite_1.AbstractSQLite {
    constructor(driver) {
        super(driver, 'txs', 
        // PK fields
        ['hash'], 
        // Fields
        [
            'hash',
            'block_number',
            'version',
            'currency',
            'comment',
            'blockstamp',
            'blockstampTime',
            'locktime',
            'received',
            'time',
            'written',
            'removed',
            'inputs',
            'unlocks',
            'outputs',
            'issuers',
            'signatures',
            'recipients',
            'output_base',
            'output_amount'
        ], 
        // Arrays
        ['inputs', 'unlocks', 'outputs', 'issuers', 'signatures', 'recipients'], 
        // Booleans
        ['written', 'removed'], 
        // BigIntegers
        [], 
        // Transient
        []);
        this.sandbox = new SandBox_1.SandBox(constants.SANDBOX_SIZE_TRANSACTIONS, this.getSandboxTxs.bind(this), (compared, reference) => {
            if (compared.output_base < reference.output_base) {
                return -1;
            }
            else if (compared.output_base > reference.output_base) {
                return 1;
            }
            else if (compared.output_amount > reference.output_amount) {
                return -1;
            }
            else if (compared.output_amount < reference.output_amount) {
                return 1;
            }
            else {
                return 0;
            }
        });
    }
    async init() {
        await this.exec('BEGIN;' +
            'CREATE TABLE IF NOT EXISTS ' + this.table + ' (' +
            'hash CHAR(64) NOT NULL,' +
            'block_number INTEGER,' +
            'locktime INTEGER NOT NULL,' +
            'version INTEGER NOT NULL,' +
            'currency VARCHAR(50) NOT NULL,' +
            'comment VARCHAR(255) NOT NULL,' +
            'time DATETIME,' +
            'inputs TEXT NOT NULL,' +
            'unlocks TEXT NOT NULL,' +
            'outputs TEXT NOT NULL,' +
            'issuers TEXT NOT NULL,' +
            'signatures TEXT NOT NULL,' +
            'recipients TEXT NOT NULL,' +
            'written BOOLEAN NOT NULL,' +
            'removed BOOLEAN NOT NULL,' +
            'PRIMARY KEY (hash)' +
            ');' +
            'CREATE INDEX IF NOT EXISTS idx_txs_issuers ON txs (issuers);' +
            'CREATE INDEX IF NOT EXISTS idx_txs_written ON txs (written);' +
            'CREATE INDEX IF NOT EXISTS idx_txs_removed ON txs (removed);' +
            'CREATE INDEX IF NOT EXISTS idx_txs_hash ON txs (hash);' +
            'COMMIT;');
    }
    getAllPending(versionMin) {
        return this.sqlFind({
            written: false,
            removed: false,
            version: { $gte: versionMin }
        });
    }
    getTX(hash) {
        return this.sqlFindOne({
            hash: hash
        });
    }
    async removeTX(hash) {
        const tx = await this.sqlFindOne({
            hash: hash
        });
        if (tx) {
            tx.removed = true;
            return this.saveEntity(tx);
        }
        return tx;
    }
    addLinked(tx, block_number, time) {
        const dbTx = DBTx.fromTransactionDTO(tx);
        dbTx.block_number = block_number;
        dbTx.time = time;
        dbTx.received = moment().unix();
        dbTx.written = true;
        dbTx.removed = false;
        dbTx.hash = tx.getHash();
        return this.saveEntity(dbTx);
    }
    addPending(dbTx) {
        dbTx.received = moment().unix();
        dbTx.written = false;
        dbTx.removed = false;
        return this.saveEntity(dbTx);
    }
    getLinkedWithIssuer(pubkey) {
        return this.sqlFind({
            issuers: { $contains: pubkey },
            written: true
        });
    }
    async getLinkedWithRecipient(pubkey) {
        const rows = await this.sqlFind({
            recipients: { $contains: pubkey },
            written: true
        });
        // Which does not contains the key as issuer
        return _.filter(rows, (row) => row.issuers.indexOf(pubkey) === -1);
    }
    getPendingWithIssuer(pubkey) {
        return this.sqlFind({
            issuers: { $contains: pubkey },
            written: false,
            removed: false
        });
    }
    getPendingWithRecipient(pubkey) {
        return this.sqlFind({
            recipients: { $contains: pubkey },
            written: false,
            removed: false
        });
    }
    insertBatchOfTxs(txs) {
        // // Be sure the recipients field are correctly updated
        DBTx.setRecipients(txs);
        const queries = [];
        const insert = this.getInsertHead();
        const values = txs.map((cert) => this.getInsertValue(cert));
        if (txs.length) {
            queries.push(insert + '\n' + values.join(',\n') + ';');
        }
        if (queries.length) {
            this.exec(queries.join('\n'));
        }
    }
    trimExpiredNonWrittenTxs(limitTime) {
        return this.exec("DELETE FROM txs WHERE NOT written AND blockstampTime <= " + limitTime);
    }
    getTransactionByExtendedHash(hash) {
        return this.query("SELECT * FROM txs WHERE hash = ? OR v4_hash = ? OR v5_hash = ?", [hash, hash, hash]);
    }
    /**************************
     * SANDBOX STUFF
     */
    getSandboxTxs() {
        return this.query('SELECT * FROM sandbox_txs LIMIT ' + (this.sandbox.maxSize), []);
    }
    getSandboxRoom() {
        return this.sandbox.getSandboxRoom();
    }
    setSandboxSize(maxSize) {
        this.sandbox.maxSize = maxSize;
    }
}
exports.TxsDAL = TxsDAL;
//# sourceMappingURL=TxsDAL.js.map