// Source file from duniter: Crypto-currency software to manage libre currency such as Ğ1
// Copyright (C) 2018  Cedric Moreau <cem.moreau@gmail.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const contacter_1 = require("./contacter");
const index_1 = require("../../../lib/common-libs/index");
const index_2 = require("../../../lib/common-libs/parsers/index");
exports.pullSandbox = async (currency, fromHost, fromPort, toHost, toPort, logger) => {
    const from = new contacter_1.Contacter(fromHost, fromPort);
    const to = new contacter_1.Contacter(toHost, toPort);
    let res;
    try {
        res = await from.getRequirementsPending(1);
    }
    catch (e) {
        // Silent error
        logger && logger.trace('Sandbox pulling: could not fetch requirements on %s', [fromHost, fromPort].join(':'));
    }
    if (res) {
        const docs = getDocumentsTree(currency, res);
        for (const identity of docs.identities) {
            await submitIdentity(identity, to);
        }
        for (const certification of docs.certifications) {
            await submitCertification(certification, to);
        }
        for (const membership of docs.memberships) {
            await submitMembership(membership, to);
        }
    }
};
exports.pullSandboxToLocalServer = async (currency, fromHost, toServer, logger, watcher = null, nbCertsMin = 1, notify = true) => {
    let res;
    try {
        res = await fromHost.getRequirementsPending(nbCertsMin || 1);
    }
    catch (e) {
        watcher && watcher.writeStatus('Sandbox pulling: could not fetch requirements on %s', [fromHost.host, fromHost.port].join(':'));
    }
    if (res) {
        const docs = getDocumentsTree(currency, res);
        for (let i = 0; i < docs.identities.length; i++) {
            const idty = docs.identities[i];
            watcher && watcher.writeStatus('Identity ' + (i + 1) + '/' + docs.identities.length);
            await submitIdentityToServer(idty, toServer, notify, logger);
        }
        for (let i = 0; i < docs.revocations.length; i++) {
            const idty = docs.revocations[i];
            watcher && watcher.writeStatus('Revocation ' + (i + 1) + '/' + docs.revocations.length);
            await submitRevocationToServer(idty, toServer, notify, logger);
        }
        for (let i = 0; i < docs.certifications.length; i++) {
            const cert = docs.certifications[i];
            watcher && watcher.writeStatus('Certification ' + (i + 1) + '/' + docs.certifications.length);
            await submitCertificationToServer(cert, toServer, notify, logger);
        }
        for (let i = 0; i < docs.memberships.length; i++) {
            const ms = docs.memberships[i];
            watcher && watcher.writeStatus('Membership ' + (i + 1) + '/' + docs.memberships.length);
            await submitMembershipToServer(ms, toServer, notify, logger);
        }
    }
};
function getDocumentsTree(currency, res) {
    const documents = {
        identities: [],
        certifications: [],
        memberships: [],
        revocations: []
    };
    for (const idty of res.identities) {
        const identity = index_1.rawer.getOfficialIdentity({
            currency,
            uid: idty.uid,
            pubkey: idty.pubkey,
            buid: idty.meta.timestamp,
            sig: idty.sig
        });
        if (idty.revocation_sig) {
            const revocation = index_1.rawer.getOfficialRevocation({
                currency,
                uid: idty.uid,
                issuer: idty.pubkey,
                buid: idty.meta.timestamp,
                sig: idty.sig,
                revocation: idty.revocation_sig
            });
            documents.revocations.push(revocation);
        }
        documents.identities.push(identity);
        for (const cert of idty.pendingCerts) {
            const certification = index_1.rawer.getOfficialCertification({
                currency,
                idty_issuer: idty.pubkey,
                idty_uid: idty.uid,
                idty_buid: idty.meta.timestamp,
                idty_sig: idty.sig,
                issuer: cert.from,
                buid: cert.blockstamp,
                sig: cert.sig
            });
            documents.certifications.push(certification);
        }
        for (const ms of idty.pendingMemberships) {
            const membership = index_1.rawer.getMembership({
                currency,
                userid: idty.uid,
                issuer: idty.pubkey,
                certts: idty.meta.timestamp,
                membership: ms.type,
                block: ms.blockstamp,
                signature: ms.sig
            });
            documents.memberships.push(membership);
        }
    }
    return documents;
}
async function submitIdentity(idty, to, logger = null) {
    try {
        await to.postIdentity(idty);
        logger && logger.trace('Sandbox pulling: success with identity \'%s\'', idty.uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitCertification(cert, to, logger = null) {
    try {
        await to.postCert(cert);
        logger && logger.trace('Sandbox pulling: success with cert key %s => %s', cert.from.substr(0, 6), cert.idty_uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitMembership(ms, to, logger = null) {
    try {
        await to.postRenew(ms);
        logger && logger.trace('Sandbox pulling: success with membership \'%s\'', ms.uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitIdentityToServer(idty, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseIdentity.syncWrite(idty);
        await toServer.writeIdentity(obj, notify);
        logger && logger.trace('Sandbox pulling: success with identity \'%s\'', obj.uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitRevocationToServer(revocation, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseRevocation.syncWrite(revocation);
        await toServer.writeRevocation(obj, notify);
        logger && logger.trace('Sandbox pulling: success with revocation \'%s\'', obj.uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitCertificationToServer(cert, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseCertification.syncWrite(cert);
        await toServer.writeCertification(obj, notify);
        logger && logger.trace('Sandbox pulling: success with cert key %s => %s', cert.from.substr(0, 6), cert.idty_uid);
    }
    catch (e) {
        // Silent error
    }
}
async function submitMembershipToServer(ms, toServer, notify, logger) {
    try {
        const obj = index_2.parsers.parseMembership.syncWrite(ms);
        await toServer.writeMembership(obj, notify);
        logger && logger.trace('Sandbox pulling: success with membership \'%s\'', ms.uid);
    }
    catch (e) {
        // Silent error
    }
}
//# sourceMappingURL=sandbox.js.map